"""
This script display or export in JSON format the staus of the Cluster
"""

import sys
import json
import eventlet
import socket
import collections
from urllib.parse import urlparse
from oslo_config import cfg
from swift.cli import recon
from swift.common.ring import Ring as swiftRing

# Initialize Swift connection
timeout = 1
swiftrecon = recon.SwiftRecon()
ring_names = ["object", "container", "account"]
swift_dir = "/etc/swift"
pool_size = 30
pool = eventlet.GreenPool(pool_size)

# Initialize script common oslo_config
CLI_OPTS = [
    cfg.BoolOpt('json', default=False, short='j',
                help='Format output as json'),
    cfg.BoolOpt('all', default=False, short='a',
                help='Active all options'),
    cfg.BoolOpt('base2', default=False,
                help='Print human-readable result in base 2 unit (kiB, MiB...)'),
]

# Initialize script cluster oslo_config
CLI_OPTS_CLUSTER = [
    cfg.BoolOpt('status', default=False, short='s',
                help='Show cluster state'),
    cfg.BoolOpt('reason', default=False, short='r',
                help='Show cluster state reason'),
    cfg.BoolOpt('ntp_drift', default=False,
                help='Show servers NTP drift'),
    cfg.BoolOpt('validate', default=False,
                help='Validate server in the swift ring'),
]

# Initialize script disks oslo_config
CLI_OPTS_DISK = [
    cfg.BoolOpt('usage', default=False, short='d',
                help='Show disk usage for object ring'),
    cfg.BoolOpt('unmounted', default=False, short='u',
                help='Show unmounted or unknow state drives'),
    cfg.BoolOpt('filling', default=False, short='f',
                help='Show disk filling for object ring'),
    cfg.IntOpt('fmin', default=10,
               help='Display Node/device filling < DISK_FMIN'),
    cfg.IntOpt('fmax', default=90,
               help='Display Node/device filling > DISK_FMAX'),
    cfg.BoolOpt('deviation', default=False,
                help='Show object ring deviation'),
]


def _prepare_config():
    """
    Prepare the oslo_config of scripts by analyse arguments
    return: the oslo_config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    CONF.register_cli_opts(CLI_OPTS_CLUSTER, group='cluster')
    CONF.register_cli_opts(CLI_OPTS_DISK, group='disk')
    return CONF


def size_suffix(size, base2=False):
    """
    Format the disk size in human readable
    param size: the size in octet
    param base2: a boolean for return result in base2, else return in base10
    return: the formatted string
    """
    units = 'iB' if base2 else 'B'
    suffixes = ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']
    kilo = 1024.0 if base2 else 1000.0
    for suffix in suffixes:
        if size < kilo:
            return "%s %s%s" % (size, suffix, units)
        size = round(size / kilo, 2)
    return "%s %s%s" % (size, suffix, units)

def std(array):
    mean = sum(array) / len(array)
    var = sum((l-mean)**2 for l in array) / len(array)
    st_dev = var ** 0.5
    return st_dev


class Cluster():
    """Class representing the entire Cluster"""

    def __init__(self, ring_names, args):
        """
        Instance method to initialize Cluster
        param ring_names: a list with the différent ring to initialize
        param args: an oslo_config object
        """
        self.rings = {}
        self.args = args
        for ring_name in ring_names:
            self.rings[ring_name] = Ring(ring_name)
        self.status = 'OK'
        self.reason = ''
        self.disks_state = {}
        self.usage, self.usage_human = {}, {}
        self.ntpdrift, self.validate = {}, {}
        self.filling, self.fmin = None, None
        self.fmax, self.deviation = None, None
        self.hosts_in_error = set()
        self.region_with_state = {'unmounted': 0, 'unknown': 0}
        self.disks_state_by_region = {}
        self.analyse()

    def analyse(self):
        """
        Instance method to analyse Cluster
        """
        # True if one options is True
        all_analyze = any(
                (
                    self.args.all,
                    self.args.json,
                    self.args.cluster.status,
                    self.args.cluster.reason
                )
        )
        # True if one options is True
        disks_analyze = any(
            (
                self.args.all,
                self.args.disk.usage,
                self.args.disk.unmounted,
                self.args.disk.filling,
                self.args.disk.fmin,
                self.args.disk.fmax,
                self.args.disk.deviation
            )
        )
        if disks_analyze or all_analyze:
            for name, ring in self.rings.items():
                # Get properties of all rings
                self.disks_state[name] = {}
                self.disks_state[name]['unmounted'] = ring.disks_unmounted()
                self.disks_state[name]['unknown'] = ring.disks_unknown()
                self.usage[name] = ring.disks_usage()
                self.usage_human[name] = ring.disks_usage_human(
                    self.args.base2)
                self.ntpdrift[name] = ring.ntpdrift()
                self.validate[name] = ring.validate()
                for name_st, ds in self.disks_state[name].items():
                    # Aggregate disks staus by region
                    for region, value in ds.items():
                        self.disks_state_by_region.setdefault(region, {})
                        for host, devices in value.items():
                            self.disks_state_by_region[region].\
                                setdefault(host, {})
                            self.disks_state_by_region[region][host] = {
                                name_st: devices}
                self.hosts_in_error.update(ring.hosts_in_error)
            for region, hosts in self.disks_state_by_region.items():
                # Increment region with unmounted or unknown disk state
                for st_name in self.region_with_state.keys():
                    if len({k: v for (k, v) in hosts.items() if st_name in v}):
                        self.region_with_state[st_name] += 1
            for state, value in self.region_with_state.items():
                if value > 1:
                    self.status = "Critical"
                    self.reason = f"{self.reason}{value} regions with {state} state drives\n"
            self.filling = self.rings['object'].disks_filling()
            self.fmin = self.rings['object'].disks_fmin(self.args.disk.fmin)
            self.fmax = self.rings['object'].disks_fmax(self.args.disk.fmax)
            self.deviation = self.rings['object'].disks_deviation()
            if self.deviation > 5 and self.status == "OK":
                self.status = "Warn"
                self.reason = "Deviation higher than 5."
        if self.status == "OK" \
                and (all_analyze or self.args.cluster.ntp_drift):
            if [n for n in self.ntpdrift.values() if n]:
                self.status = "Warn"
                self.reason = "Some servers have ntpdrift"
        if self.status == "OK" and (all_analyze or self.args.cluster.validate):
            if [n for n in self.ntpdrift.values() if n]:
                self.status = "Warn"
                self.reason = "Some servers cannot be validate"

    def display(self):
        """
        Instance method to display status of Cluster
        """
        if self.args.all:
            print(f"Cluster status: {self.status}")
            if self.reason:
                print(f"Cluster state reason: {self.reason}\n")
            if self.hosts_in_error:
                print("WARNING, hosts in error:")
                for h in sorted(self.hosts_in_error):
                    print(f"\t{h}")
                print('')
            self.display_unmounted()
            self.display_ntpdrift()
            self.display_validate()
            self.display_usage()
            self.display_filling()
            self.display_deviation()
        else:
            if self.args.cluster.status:
                print(f"Cluster status: {self.status}")
            if self.args.cluster.reason and self.reason:
                print(f"Cluster state reason: {self.reason}")
            if len(self.hosts_in_error) > 0:
                print("WARNING, hosts in error:")
                for h in sorted(self.hosts_in_error):
                    print(f"\t{h}")
                print('')
            if self.args.disk.unmounted:
                self.display_unmounted()
            if self.args.cluster.ntp_drift:
                self.display_ntpdrift()
            if self.args.cluster.validate:
                self.display_validate()
            if self.args.disk.usage:
                self.display_usage()
            if self.args.disk.filling:
                self.display_filling()
            if self.args.disk.deviation:
                self.display_deviation()

    def display_unmounted(self):
        """
        Instance method to display disks in unmounted or unknown state
        """
        rs = self.region_with_state
        print(f"Region with umounted/unknown state drives: {rs['unmounted']}/ {rs['unknown']}")
        if self.region_with_state['unmounted'] \
                or self.region_with_state['unknown']:
            print("Unmounted/unknown state disks output per region:")
            for region, nodes in self.disks_state_by_region.items():
                txt = ''
                for node, v in sorted(nodes.items()):
                    unmounted = ','.join(sorted(v['unmounted']
                                         if 'unmounted' in v else []))
                    unknown = ','.join(sorted(v['unknown']
                                              if 'unknown' in v else []))
                    txt = f"{txt}\t{node}: ({unmounted})/({unknown})\n"
                if txt:
                    print(f"    {region}:{txt}")
        print('')

    def display_ntpdrift(self):
        """
        Instance method to display the servers with Ntp drift in the cluster
        """
        print("Ntp drift:")
        for name, ntp in self.ntpdrift.items():
            print(f"  {name}: {ntp}")
        print('')

    def display_validate(self):
        """
        Instance method to display the invalidate servers in the cluster
        """
        print("Validate server output:")
        for name, val in self.validate.items():
            print(f"  {name}: {val}")
        print('')

    def display_usage(self):
        """
        Instance method to display the disks usage in human readable
        """
        print("Disk usage human readable output:")
        for ring, usage in self.usage_human.items():
            print(f"  {ring}: {usage['used']} used on {usage['size']} {usage['percent']}%")
        print('')

    def display_filling(self):
        """
        Instance method to display the disks filling in pseudo graph
        for object ring
        """
        print("Disk filling for object ring:")
        tpercent, tvalue = 0, 0
        # Calcul number of disks by representing by #
        div = int(max(self.filling.values()) / 100) + 1
        for k, v in self.filling.items():
            # Shows disk fill distribution
            nb = round(v / div)
            print(f"\t{k:2}% -> {v:<5d}{nb * '#'}")
            tpercent += (k * v)
            tvalue += v
        print("Average disk filling for object ring:")
        print(f"\t{tpercent/tvalue:2.2f}% -> {tvalue}")
        # Show disks with less than fmin padding
        if self.fmin and self.fmin.items():
            print(f"\nList of Disk filling < {self.args.disk.fmin} for object ring:")
            for k, v in sorted(self.fmin.items(), key=lambda x:x[1]):
                print(f"\t {k} -> {round(v,2)}%")
        # Show disks with more than fmax padding
        if self.fmax and self.fmax.items():
            print(f"\nList of Disk filling > {self.args.disk.fmax} for object ring:")
            for k, v in sorted(self.fmax.items(), key=lambda x: x[1], reverse=True):
                print(f"\t {k} -> {round(v,2)}%")

    def display_deviation(self):
        """
        Instance method to display the deviation filling for object ring
        """
        print(f"\nDeviation for object ring: {round(self.deviation,2)}")

    def toJson(self):
        """
        Instance method to export all status in json format
        """
        formated_output = {
            "status": self.status,
            "hosts_in_error": self.hosts_in_error,
            "reason": self.reason,
            "unmounted_per_region": self.unmounted,
            "unknowstate_per_region": self.unknown,
            "ntp_drift": self.ntpdrift,
            "validation": self.validate,
            "disk_usage": self.usage,
            "disk_usage_human": self.usage_human,
            "disk_filling": self.filling,
            "df_min": self.fmin,
            "df_max": self.fmax,
            "deviation": self.deviation,
        }
        print(json.dumps(formated_output))


class Ring():
    """Class representing a ring of  Cluster"""

    def __init__(self, ring_name):
        """
        Instance method to initialize Cluster
        param ring_name: the ring name
        """
        self.ring = swiftRing(swift_dir, ring_name=ring_name)
        self.ring_name = ring_name
        self.hosts = {}
        self.hosts_in_error = set()
        self.host_port = set()
        self.load_hosts()
        self.load_devices()

    def load_hosts(self):
        """
        Instance method to load information for host
        """
        d = {}
        devs = [d for d in self.ring.devs if d]
        for dev in devs:
            self.hosts[dev['ip']] = Host.create(dev)
            d[dev['ip']] = dev['port']
        self.host_port = set(d.items())

    def load_devices(self):
        """
        Instance method to load informations for all devices
        and assign to hosts
        """
        scout = recon.Scout(recon_type="diskusage",
                            timeout=timeout,
                            suppress_errors=True)
        for url, response, status, _ts_start, _ts_end in \
                pool.imap(scout.scout, self.host_port):
            ip = urlparse(url).hostname
            if status == 200:
                for res in response:
                    self.hosts[ip].devices_by_name[
                        res['device']].load_datas(res)
            else:
                self.hosts[ip].error = True

    def add_host_in_error(self, url):
        """
        Instance method to add host in error to the list
        param url: the host url
        """
        ip = urlparse(url).hostname
        try:
            hostname = socket.gethostbyaddr(ip)[0]
        except Exception:
            hostname = ip
        self.hosts_in_error.add(hostname)

    def ntpdrift(self):
        """
        Instance method search host with ntp drift
        return: list of hosts with ntp drift
        """
        jitter = abs(0.0)
        scout = recon.Scout(recon_type="time",
                            timeout=timeout,
                            suppress_errors=True)
        ntpdrift_data = []
        swiftrecon._ptime
        for url, ts_remote, status, ts_start, ts_end \
                in pool.imap(scout.scout, self.host_port):
            if status != 200:
                self.add_host_in_error(url)
                continue
            # Must round as request take some ms
            ts_remote = round(ts_remote, 1)
            ts_start = round(ts_start, 1)
            ts_end = round(ts_end, 1)
            if ts_remote + jitter < ts_start or ts_remote - jitter > ts_end:
                hostname = self.hosts[urlparse(url).hostname].hostname
                if hostname not in ntpdrift_data:
                    ntpdrift_data.append(hostname)
                continue
        return ntpdrift_data

    def validate(self):
        """
        Instance method to load invalid hosts
        return: list of invalid hosts
        """
        errors = []
        validate_data = []
        scout = recon.Scout(recon_type="server_type_check",
                            timeout=timeout,
                            suppress_errors=True)
        for url, response, status in pool.imap(scout.scout_server_type,
                                               self.host_port):
            if status == 200:
                if response != self.ring_name + "-server":
                    errors.append(url)
            else:
                self.add_host_in_error(url)
        for url in errors:
            hostname = self.hosts[urlparse(url).hostname].hostname
            if hostname not in validate_data:
                validate_data.append(hostname)
        return validate_data

    def disks_usage(self):
        """
        Instance method to load disk usage of all hosts
        """
        used = 0
        avail = 0
        size = 0
        for h in self.hosts.values():
            for d in h.devices_by_port.values():
                used += d.used if isinstance(d.used, int) else 0
                avail += d.avail if isinstance(d.avail, int) else 0
                size += d.size if isinstance(d.size, int) else 0
        percent = round(100 * used / size, 2)
        return {'used': used, 'avail': avail, 'size': size, 'percent': percent}

    def disks_usage_human(self, base2):
        """
        Instance method to load disk usage of all hosts
        return: dict with used, avail, total size and percent filling
        """
        du = self.disks_usage()
        return {'used': size_suffix(du['used'], base2),
                'avail': size_suffix(du['avail'], base2),
                'size': size_suffix(du['size'], base2),
                'percent': du['percent']}

    def disks_filling(self):
        """
        Instance method to load filling of all hosts
        return: dict with distribution of filling
        """
        freq = {}
        for h in self.hosts.values():
            for d in h.devices_by_port.values():
                if d.filling in freq:
                    freq[d.filling] += 1
                else:
                    freq[d.filling] = 1
        freq = collections.OrderedDict(sorted(freq.items()))
        return freq

    def disks_deviation(self):
        """
        Instance method to load deviation of filling
        return: the deviation
        """
        filling = []
        for h in self.hosts.values():
            filling += [d.filling for d in h.devices_by_port.values()]
        return (std(filling))

    def disks_fmin(self, dfmin):
        """
        Instance method to load devices filling less than dfmin
        param dfmin: the minimum filling to detect
        return: dict of host/device with filling less than dfmin
        """
        stats = {}
        for h in self.hosts.values():
            for d in h.devices_by_name.values():
                try:
                    usage = 100 * d.used / d.size
                    if usage < dfmin:
                        stats[f"{h.hostname}/{d.name}"] = usage
                except Exception:
                    pass
        return stats

    def disks_fmax(self, dfmax):
        """
        Instance method to load devices filling more than dfmax
        param dfmax: the maximum filling to detect
        return: dict of host/device with filling more than dfmax
        """
        stats = {}
        for h in self.hosts.values():
            for d in h.devices_by_name.values():
                try:
                    usage = 100 * d.used / d.size
                    if usage > dfmax:
                        stats[f"{h.hostname}/{d.name}"] = usage
                except Exception:
                    pass
        return stats

    def disks_unmounted(self):
        """
        Instance method to load devices unmoiunted or unknow state
        return: dict with unmounted or unknow state devices
        """
        unmounted = {}
        for h in self.hosts.values():
            unmounted.setdefault(h.region, {})
            for d in h.devices_by_port.values():
                if not d.mounted:
                    unmounted[h.region].setdefault(h.hostname, [])
                    unmounted[h.region][h.hostname].append(d.name)
        return unmounted

    def disks_unknown(self):
        unknown = {}
        for h in self.hosts.values():
            unknown.setdefault(h.region, {})
            for d in h.devices_by_port.values():
                if d.unknown:
                    unknown[h.region].setdefault(h.hostname, [])
                    unknown[h.region][h.hostname].append(d.name)
        return unknown


class Host():
    """Class representing an host"""
    List = {}

    def create(device):
        """
        Class method to create a device and attach it to an host instance
        return: instance of device
        """
        if device['ip'] not in Host.List:
            Host.List[device['ip']] = Host(device)
        else:
            Host.List[device['ip']].add_device(device)
        return Host.List[device['ip']]

    def __init__(self, device):
        """
        Instance method to initialize Host with a device
        param device: a device to attach to the new host
        """
        self.ip = device['ip']
        self.hostname = socket.gethostbyaddr(self.ip)[0]
        self.zone = device['zone']
        self.region = device['region']
        self.devices_by_name = {}
        self.devices_by_port = {}
        self.add_device(device)
        self.error = False

    def add_device(self, device):
        """
        Instance method to add device to the host
        param device: device to attach to the Host instance
        """
        d = Device(device)
        self.devices_by_name[d.name] = d
        self.devices_by_port[d.port] = d


class Device():
    """Class representing a device"""

    def __init__(self, device):
        """
        Instance method to initialize Device
        param device: device to initialize
        """
        self.port = device['port']
        self.name = device['device']
        self.weight = device['weight']
        self.used = 0
        self.avail = 0
        self.size = 0
        self.filling = 0
        self.mounted = False
        self.unknown = False

    def load_datas(self, datas):
        """
        Instance method to load data off device
        param device: device to initialize
        """
        self.used = datas['used']
        self.avail = datas['avail']
        self.size = datas['size']
        if isinstance(datas['mounted'], bool) and datas['mounted']:
            self.mounted = datas['mounted']
        elif datas['mounted'] is None:
            self.unknown = True
        try:
            self.filling = round(100 * self.used / self.size)
        except Exception:
            pass


def main():
    """
    Entry point for the script
    """
    args = _prepare_config()
    args(sys.argv[1:])
    cluster = Cluster(ring_names, args)
    cluster.display()


if __name__ == "__main__":
    sys.exit(main())



"""
This script check if rsync is disable for disks full
and can send to graphite (by default) or list to stdout
"""
import sys
import socket
import time
import configparser
from oslo_config import cfg
from subprocess import Popen, PIPE

with open('/etc/rsyncd.conf') as f:
    file_content = '[globale]\n' + f.read()

default_cluster = socket.gethostname().split('.')[0].split('-')[0]
# Initialize script common oslo_config
CLI_OPTS = [
    cfg.StrOpt('server', short='s',
               help='Graphite server to send data'),
    cfg.IntOpt('port', default=2003, short='p',
               help='Graphite server port to send data'),
    cfg.StrOpt('cluster', default=default_cluster, short='c',
               help='Cluster name'),
    cfg.BoolOpt('list', short='l',
                help='List of disk with rsync disabled'),
]


def _prepare_config():
    """
    Prepare the oslo_config of scripts by analyse arguments
    return: the oslo_config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    return CONF


def netcat(server, port, content):
    """
    Send data to graphite server
    """
    p = Popen(['/usr/bin/ncat', server, str(port)], stdin=PIPE)
    p.communicate(input=content.encode())


def print_list(disable):
    """
    Print a list of disk with rsync disable by type of server
    param disable: list of module rsync disable (ServerType_Disk)
    """
    disabled_disks = {}
    for k in disable:
        type, disk = k.strip().split('_')
        disabled_disks.setdefault(type, [])
        disabled_disks[type].append(disk)
    for k, v in disabled_disks.items():
        print(f"Rsync disks disable for type {k}:")
        for disk in v:
            print(f"\t{disk}")


def send_metric(args, disable):
    """
    Send metric to graphite
    param args: args of the script
    param disable: list of module rsync disable (ServerType_Disk)
    """
    if not args.server:
        print("Graphite server add ress not define")
        sys.exit(1)
    host = socket.gethostname().split('.')[0]
    date = time.time()
    netcat(args.server, args.port,
           f"{args.cluster}.{host}.rsync_disable {len(disable)} {date}\n")


def main():
    """
    Entry point for the script
    """

    args = _prepare_config()
    try:
        args(sys.argv[1:])
    except cfg.RequiredOptError as E:
        print(E)
        args.print_usage()
        return 1

    config_parser = configparser.RawConfigParser()
    config_parser.read_string(file_content)
    disable = [k for k, v in config_parser._sections.items() if k != 'globale'
               and int(v['max connections']) < 0]
    if args.list:
        print_list(disable)
    else:
        send_metric(args, disable)


if __name__ == "__main__":
    sys.exit(main())



import os
import sys
from oslo_config import cfg
from swift.common.ring import Ring as swiftRing
from random import randint
import json
from socket import gethostbyaddr, herror
import re

# Initialize Arguments
CLI_OPTS = [
    cfg.IntOpt('nb_tests', short='n', default=500,
               help='Number of partition to test'),
    cfg.BoolOpt('json', short='j',
                help='Show full differences in json formatted string'),
    cfg.StrOpt('pattern', short='p', default=r".*-(?P<idx>\w+)$",
               help='Pattern to extract idx of swiftstore'),
    cfg.StrOpt('Ring1', positional=True, required=True,
               help='New Ring File to compare'),
    cfg.StrOpt('Ring2', positional=True, required=True,
               help='Old Ring File to compare'),
]


def _prepare_config():
    """
    Prepare the oslo_config of scripts by analyse arguments
    return: the oslo_config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    return CONF


def get_short_hostname(ip_address):
    """
    Return short hostname for an IP address
    param ip_address: IP address to resolve
    return: short hostname
    """
    try:
        hostname = gethostbyaddr(ip_address)[0].split('.')[0]
        return hostname
    except herror:
        return "Unknown host"


def extract_host_idx(hostname, pattern):
    """
    Extract an index from a swiftstore name
    param hostname: hostname for exctract index
    param pattern: a regexp pattern to extract index
    return: index
    """
    m = re.match(pattern, hostname)
    try:
        return int(m.group('idx'))
    except Exception:
        return -1


class Ring():
    """Class representing an Ring"""

    def __init__(self, file):
        """
        Instance method to initialize a Ring
        param file: file for the ring (with ring.gz extension)
        """
        self.ring_name = os.path.basename(file)[:-len('.ring.gz')]
        self.ring = swiftRing(file)
        if '-' in self.ring_name and self.ring_name.startswith('object'):
            self.loc = 'objects-' + self.ring_name.split('-', 1)[1]
        else:
            self.loc = self.ring_name + 's'

    def get_nodes(self, part):
        """
        Instance method that return list of node for a partition
        param part: partition to test
        return: list of node and device that have the partition
        """
        part = int(part)
        primary_nodes = self.ring.get_part_nodes(part)
        return primary_nodes

    def get_md5(self):
        """
        Instance method that return de md5 for the ring
        return: md5
        """
        return self.ring.md5

    def get_devs(self):
        """
        Get device in ring
        return: dict with device id and (hostname, device_name)
        """
        devs = [d for d in self.ring.devs if d]
        devs_ip = {dev['id']: (get_short_hostname(dev['ip']), dev['device'])
                   for dev in devs}
        return devs_ip

    def get_parts(self):
        """
        Get parts in ring
        return: dict of list of partitions by device id
        """
        partitions = {}
        for replica in self.ring._replica2part2dev_id:
            for partition, device in enumerate(replica):
                partitions.setdefault(device, set())
                partitions[device].add(partition)
        return partitions

    def partitions_count(self):
        """
        Instance method that return number of partition for the ring
        return: number of partitions
        """
        return len(self.ring._replica2part2dev_id[0])

    def compare_with(self, other_r, part):
        """
        Instance method that compare this ring with another
        If more than 1 difference, display an Error Message and exit
        param other_r: the other ring to compare
        param part: the partition to test
        return: the count of difference and nodes/device affected
        """
        my_nodes = self.get_nodes(part)
        other_n = other_r.get_nodes(part)
        diff_count = 0
        diff = {}
        for i in range(3):  # 3 because replica 3 in both rings
            diff_items = {k: my_nodes[i][k] for k in my_nodes[i]
                          if k in other_n[i] and
                          my_nodes[i][k] != other_n[i][k]}
            if diff_items and not (len(diff_items) == 1
               and "weight" in diff_items):
                diff[self.ring_name] = {'ip': my_nodes[i]['ip'],
                                        'port': my_nodes[i]['port'],
                                        'device': my_nodes[i]['device']}
                diff[other_r.ring_name] = {'ip': other_n[i]['ip'],
                                           'port': other_n[i]['port'],
                                           'device': other_n[i]['device']}
                diff_count += 1
                if diff_count > 1:
                    print("[ERROR] more than 1 differences for one partition, \
                           new ring is not valid.", file=sys.stderr)
                    sys.exit(1)
        return diff_count, diff

    def get_in_out(self, other_r):
        """
        Instance method to extract partitions in/out devices between 2 rings
        param other_r: the other ring to compare
        return: 2 dicts for partitions In and Out of each devices
        """
        parts_in = {}
        parts_out = {}
        self_parts = self.get_parts()
        other_r_parts = other_r.get_parts()
        list_host = list(self_parts.keys()) + list(other_r_parts.keys())
        for k in list_host:
            parts_in.setdefault(k, 0)
            parts_out.setdefault(k, 0)
            if k not in other_r_parts:
                other_r_parts[k] = set()
            if k not in self_parts:
                self_parts[k] = set()
            parts_in[k] += len(self_parts[k] - other_r_parts[k])
            parts_out[k] += len(other_r_parts[k] - self_parts[k])
        return parts_in, parts_out


def main():
    """
    Entry point for the script
    """
    args = _prepare_config()
    try:
        args(sys.argv[1:])
    except cfg.RequiredOptError as E:
        print(E)
    # Get informations for 2 rings
    ring1 = Ring(args.Ring1)
    ring2 = Ring(args.Ring2)
    md5_1, md5_2 = None, None
    try:
        md5_1 = ring1.get_md5()
        md5_2 = ring2.get_md5()
    except Exception:
        pass
    partitions_count = ring1.partitions_count()

    diffs = {}
    nb_diffs = 0
    # Compare the rings for args.nb_test random partitions
    for i in range(args.nb_tests):
        part = randint(1, partitions_count)
        nb_diff, diff = ring1.compare_with(ring2, part)
        nb_diffs += nb_diff
        if diff:
            diffs[f"Partition {part}"] = diff

    # Extract In and Out partitions by devices
    parts_in, parts_out = ring1.get_in_out(ring2)

    # Show result
    print()
    if args.json:
        diffs[f"Ring {ring1.ring_name} md5"] = md5_1
        diffs[f"Ring {ring2.ring_name} md5"] = md5_2
        diff['Partitions IN'] = parts_in
        diff['Partitions OUT'] = parts_out
        print(json.dumps(diffs))
    else:
        if md5_1 == md5_2 and md5_1 is not None and md5_2 is not None:
            print("md5 for two are identical !")
            sys.exit(0)
        else:
            print(f"Ring {ring1.ring_name} md5 : {md5_1}")
            print(f"Ring {ring2.ring_name} md5 : {md5_2}")
        if nb_diffs:
            print(f"Number of differences between {ring1.ring_name} \
                    and {ring2.ring_name}: {nb_diffs}/{args.nb_tests} \
                    partitions tested")
        else:
            print(f"No differences find between rings {ring1.ring_name} \
and {ring2.ring_name} , you can re-run the script for confirmation")
            return 0

    print()
    devs = ring1.get_devs()
    devs.update(ring2.get_devs())
    sum_in = 0
    hosts_move = {}
    # Group partitions IN and OUT by host
    for k, v in parts_in.items():
        idx = extract_host_idx(devs[k][0], args.pattern)
        hosts_move.setdefault(idx, {'name': devs[k][0], 'in': 0, 'out': 0})
        sum_in += v
        hosts_move[idx]['in'] += v
    sum_out = 0
    for k, v in parts_out.items():
        idx = extract_host_idx(devs[k][0], args.pattern)
        hosts_move.setdefault(idx, {'name': devs[k][0], 'in': 0, 'out': 0})
        sum_out += v
        hosts_move[idx]['out'] += v

    print(f"Total partitions moved: {sum_in} In, {sum_out} Out.")
    maxi = 0
    for k, v in hosts_move.items():
        maxi = v['in'] if v['in'] > maxi else maxi
        maxi = v['out'] if v['out'] > maxi else maxi

    div = (maxi / 50)
    len_idx = int(max(list(hosts_move.keys())) / 4) + 1
    format_in = [0] * (50 * len_idx)
    format_out = [0] * (50 * len_idx)
    for k, v in sorted(hosts_move.items()):
        idx = int((k - 1) / 4) if k >= 0 else -1
        reste = 2 ** ((k - 1) % 4)
        nb_in = int(v['in'] / div)
        nb_out = int(v['out'] / div)
        for i in range(0, 50):
            format_in[idx * 50 + i] += (reste if nb_in > i else 0)
            format_out[idx * 50 + i] += (reste if nb_out > i else 0)
    # For grouping graph line by 4 hosts
    char_graph = [' ', '⠉', '⠒', '⠛', '⠤', '⠭', '⠶', '⠿',
                  '⣀', '⣉', '⣒', '⣛', '⣤', '⣭', '⣶', '⣿']
    print()
    pr = f"{'Hosts':28}: \x1b[1;31;40m{'IN max: ' + str(int(div * 50)):<50}"
    pr = f"{pr}\x1b[0;37;40m|\x1b[1;32;40m"
    pr = f"{pr}{'OUT max: ' + str(int(div * 50)):>50}\x1b[0;37;40m"
    print(pr)
    print('-' * 131)
    for idx in range(0, len_idx):
        try:
            hosts = f"{hosts_move[1 + idx * 4]['name']}…{idx * 4 + 4}"
        except Exception:
            if -1 in hosts_move.keys():
                hosts = f"{hosts_move[-1]['name']}"
            else:
                hosts = "Hostname unknown"
        print(f"{hosts:28}: ", end='')
        print('\x1b[1;31;40m', end='')
        for i in range(49, -1, -1):
            try:
                print(char_graph[format_in[idx * 50 + i]], end='')
            except Exception:
                print(' ' * 50, end='')
        print('\x1b[0;37;40m|\x1b[1;32;40m', end='')
        for i in range(0, 50):
            try:
                print(char_graph[format_out[idx * 50 + i]], end='')
            except Exception:
                print(' ' * 50, end='')
        print('\x1b[0;37;40m')
    sys.exit(nb_diffs)


if __name__ == "__main__":
    sys.exit(main())



"""
This script check if a    device has dark datas
By default it display it and it can clean it
"""
import os
import sys
from oslo_config import cfg
from swift.common.ring import Ring
from socket import gethostname, gethostbyname
import shutil

# Initialize ring
try:
    ring = Ring('/etc/swift/object.ring.gz')
except Exception as e:
    sys.exit(e)

# Initialize script arguments
CLI_OPTS = [
    cfg.BoolOpt('clean-dark-parts', help='Delete dark partitions on disk'),
    cfg.BoolOpt('yes-i-really-mean-it', help='To really delete partitions'),
]


def _prepare_config():
    """
    Prepare the oslo config
    return: oslo config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    return CONF


def get_ring_devs(ip):
    """
    Get device in ring for this server
    param ip: IP address of this server
    return: dict with device id and its name
    """
    devs = [d for d in ring.devs if d]
    devs_filter = [d for d in devs if d['ip'] == ip]
    if devs_filter:
        devs_ip = {d['id']: d['device'] for d in devs_filter}
    return devs_ip


def get_ring_parts(devs):
    """
    Get part in ring for a device
    param devs: list of devices in a server
    return: dict of list of partitions by device
    """
    partitions = {}
    for replica in ring._replica2part2dev_id:
        for partition, device in enumerate(replica):
            if device in devs:
                partitions.setdefault(devs[device], [])
                partitions[devs[device]].append(partition)
    return partitions


def get_dark_datas(ip, path, partitions):
    """
    Get dark datas on a device
    param ip: IP address of this server
    param path: path for search partitions on a Device
    params partitions: partitions on the ring for a device
    return: list of dark partitions
    """
    try:
        parts = [int(d) for d in os.listdir(path)
                 if os.path.isdir(os.path.join(path, d))]
    except Exception:
        print("Explore {path} -> Disk unmounted !", file=sys.stderr)
        return []
    diff = set(parts)-set(partitions)
    darks = list(diff)
    for dark in darks:
        for hand in ring.get_more_nodes(dark):
            if hand['handoff_index'] >= 3:
                break
            if ip == hand['ip']:
                darks.remove(dark)
                break
    return darks


def main():
    """
    Entry point for the script
    """
    CONF = _prepare_config()
    CONF(sys.argv[1:])
    if CONF.clean_dark_parts and not CONF.yes_i_really_mean_it:
        print('You should confirme delete by --yes-i-really-mean-it !',
              file=sys.stderr)
        return 1
    ip = gethostbyname(gethostname())
    devs = get_ring_devs(ip)
    parts = get_ring_parts(devs)
    for dev, parts in parts.items():
        path = os.path.join('/srv/node', dev, 'objects')
        darks = get_dark_datas(ip, path, parts)
        if CONF.clean_dark_parts and CONF.yes_i_really_mean_it:
            for dark in darks:
                dark_part = os.path.join(path, dark)
                print(f"Remove {dark_part} !")
                try:
                    shutil.rmtree(dark_part)
                except OSError as e:
                    print("Error: %s - %s." % (e.filename, e.strerror))
        elif darks:
            print('    ', dev, '>>>', ', '.join([str(d) for d in darks]))
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
This script test if a Swift cluster ( or a swiftsore server)
is under replication
"""

import sys
import socket
import eventlet
from oslo_config import cfg
from swift.cli import recon
from swift.common.ring import RingBuilder
from urllib.parse import urlparse

# Initialize Swift connection
swiftrecon = recon.SwiftRecon()
pool_size = 30
pool = eventlet.GreenPool(pool_size)

CLI_OPTS = [
    cfg.StrOpt('type', short='t', choices=['object', 'container', 'account'],
               default='object', help='Choose de type of server'),
]


def _prepare_config():
    """
    Prepare the oslo_config of scripts by analyse arguments
    return: the oslo_config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    return CONF


def analyze_ring():
    """
    Analyze the ring for retrieve informations about host or cluster
    If script running in a switstore, he check only this server,
    otherwise, he check all cluster
    return: list off tuble with ip and port, total partitions on this hosts
            parttion to be emptied ans filled
    """
    hosts = {}
    parts = 0
    filling = {}
    emptying = {}
    devs = [d for d in builder.devs if d]
    # If server have disk, use it, else use all swiftstore
    ip = socket.gethostbyname(socket.gethostname())
    devs_ip = [d for d in devs if d['ip'] == ip]
    if devs_ip:
        devs = devs_ip
    for h in devs:
        hosts[h['ip']] = h['port']
        parts += h['parts']
        if h['parts_wanted'] < -builder.parts:
            # If emtying swiftstore, parts-wanted is
            # (builder.parts * builder.replicas) -
            # total.parts to rebalance + device parts to rebalance
            # So is always < -builder.parts
            emptying.setdefault(h['ip'], 0)
            emptying[h['ip']] += h['parts']
        elif h['parts_wanted'] < 0:
            # If rebalance after add swifstore, this juste parts number
            # in swiftstore that they are emptied
            emptying.setdefault(h['ip'], 0)
            emptying[h['ip']] -= h['parts_wanted']
        elif h['parts_wanted'] > 0:
            # In all case this number of part of device that theyr are filled
            filling.setdefault(h['ip'], 0)
            filling[h['ip']] += h['parts_wanted']
    return set(hosts.items()), parts, filling, emptying


def get_attempted(hosts):
    """
    Analyze all hosts
    return: tuple with number of replication in progress
    """
    scout = recon.Scout(recon_type=f"replication/{args.type}",
                        suppress_errors=True)
    attempted = {}
    for url, response, status, _ts_start, _ts_end \
            in pool.imap(scout.scout, hosts):
        if status == 200:
            ip = urlparse(url).hostname
            attempted[ip] = response['replication_stats']['attempted']
    return sum(attempted.values())


def main():
    """
    Entry point for the script
    """
    global builder, args
    args = _prepare_config()
    try:
        args(sys.argv[1:])
    except cfg.RequiredOptError as E:
        print(E)
        args.print_usage()
        return 1

    try:
        builder = RingBuilder.load(f"/etc/swift/{args.type}.builder")
    except Exception as e:
        sys.exit(e)

    hosts, parts, filling, emptying = analyze_ring()
    total_emptying = sum(emptying.values())
    total_filling = sum(filling.values())
    percent_emptying = total_emptying / parts
    percent_filling = total_filling / parts
    if total_emptying or total_filling:
        print("Dispersion for the cluster is not optimal !")
        print(f"{total_filling} parts remains to be filled \
              ({percent_filling:.2%}) on {len(filling)} hosts")
        print(f"{total_emptying} parts remains to be emptied \
               ({percent_emptying:.2%}) on {len(emptying)} hosts")
        print("You can re-run 'swift-ring-builder object.builder rebalance' \
               AFTER the replication is OK !")
        print("=" * 80)

    attempted = get_attempted(hosts)
    inProgress = abs(attempted - parts)
    perCent = 1 - inProgress / parts
    if inProgress:
        print(f"Replication in progress: {inProgress} parts to sync \
              ({perCent:.2%} Ok) ")
    else:
        print("Replication OK")


if __name__ == "__main__":
    sys.exit(main())



"""
This script send replication progress to graphite
"""
from oslo_config import cfg
import os
from os.path import join
import sys
import socket
import datetime
import time
import random
from subprocess import Popen, PIPE
from swift.common.ring import Ring
# from swift.obj.replicator import Stats
# from collections import defaultdict
from six.moves.configparser import ConfigParser

units = {'h': 3600, 'm': 60, 's': 1}
default_cluster = socket.gethostname().split('.')[0].split('-')[0]
# Initialize script common oslo_config
CLI_OPTS = [
    cfg.StrOpt('server', short='s', required=True,
               help='Graphite server to send data'),
    cfg.IntOpt('port', default=2003, short='p',
               help='Graphite server port to send data'),
    cfg.StrOpt('cluster', default=default_cluster, short='c',
               help='Cluster name'),
]


def _prepare_config():
    """
    Prepare the oslo_config of scripts by analyse arguments
    return: the oslo_config object
    """
    CONF = cfg.ConfigOpts()
    CONF.register_cli_opts(CLI_OPTS)
    return CONF


def netcat(server, port, content):
    """
    Send data to graphite server
    """
    p = Popen(['/usr/bin/ncat', server, str(port)], stdin=PIPE)
    p.communicate(input=content.encode())


def build_replication_jobs(ring, ips):
    """
    Helper function for collect_jobs to build jobs for replication
    """
    jobs = []
    # ring2 = ring
    # ring2.all_devs_info = set()
    # ring2.stats_for_dev = defaultdict(Stats)
    # ring2.all_devs_info.update(
    #     [(dev['replication_ip'], dev['device'])
    #         for dev in ring.devs if dev])
    data_dir = "/srv/node"
    # found_local = False
    for local_dev in [device for device in ring.devs if device['ip'] == ips]:
        # found_local = True
        dev_path = join(data_dir, local_dev['device'])
        obj_path = join(dev_path, "objects")
        for partition in os.listdir(obj_path):
            if (partition.startswith('auditor_status_') and
                    partition.endswith('.json')):
                # ignore auditor status files
                continue

            part_nodes = None
            try:
                job_path = join(obj_path, partition)
                part_nodes = ring.get_part_nodes(int(partition))
                nodes = [node for node in part_nodes
                         if node['id'] != local_dev['id']]
                jobs.append(
                    dict(path=job_path,
                         device=local_dev['device'],
                         nodes=nodes,
                         partition=partition))
            except ValueError:
                continue
    # if not found_local:
    #     print("Can't find with ips %s and with port in ring file",
    #           ", ".join(ips))
    return jobs


def collect_jobs(ring):
    """
    Returns a sorted list of jobs (dictionaries) that specify the
    partitions, nodes, etc to be rsynced.
    """
    jobs = []
    swift_conf = ConfigParser()
    with open("/etc/swift/object-server.conf",
              encoding='latin1') as swift_conf_file:
        swift_conf.read_file(swift_conf_file)
    ips = swift_conf.get('DEFAULT', 'bind_ip')
    jobs += build_replication_jobs(ring, ips)
    return jobs


def main():
    """
    Entry point for the script
    """
    args = _prepare_config()
    try:
        args(sys.argv[1:])
    except cfg.RequiredOptError as E:
        print(E)
        args.print_usage()
        sys.exit(1)
    name = socket.gethostname().split('.')[0]
    date = datetime.datetime.now().timestamp()
    try:
        ring = Ring("/etc/swift/object.ring.gz")
        jobs = collect_jobs(ring)
        repl_jobs = []
        for job in jobs:
            nodes_list = []
            nodes_list.append(job['partition'])
            nodes_list.append(job['path'])
            for node in job['nodes']:
                nodes_list.append(node['replication_ip'])
                nodes_list.append(node['device'])
                if len(nodes_list) == 8:
                    repl_jobs.append(nodes_list)
        netcat(args.server, args.port,
               f"{args.cluster}.{name}.replication.handoff "
               f"{len(repl_jobs)} {date}\n")
    except Exception:
        return

    try:
        n_async = 0
        for dir in os.listdir('/srv/node'):
            for root, dirs, files in os.walk(os.path.join('/srv/node',
                                                          dir,
                                                          'async_pending')):
                n_async += len(files)
        netcat(args.server, args.port,
               f"{args.cluster}.{name}.async_pending_total {n_async} {date}\n")
    except Exception:
        return
    line_log = ''
    try:
        if os.path.isfile("/var/log/swift/swift-object-replicator.log"):
            log_path = "/var/log/swift/swift-object-replicator.log"
        elif os.path.isfile("/var/log/swift/swift-object.log"):
            log_path = "/var/log/swift/swift-object.log"
        else:
            log_path = "/var/log/swift/object.log"

        with open(log_path, 'r') as file:
            for line in file:
                if line.endswith('remaining)\n'):
                    line_log = line
    except Exception:
        return
    data = {}
    elems = line_log.split()
    try:
        date = datetime.datetime.strptime(
                        elems[0], '%Y-%m-%dT%H:%M:%S.%f%z').timestamp()
        name = elems[1]
        done, total = elems[3].split('/')
        data['parts_remaining'] = int(total) - int(done)
        data['duration'] = elems[8].split('.')[0]
        data['remaining'] = int(elems[10][:-1]) * units[elems[10][-1]]
        time.sleep(random.randint(0, 59))
        for k, v in data.items():
            netcat(args.server, args.port,
                   f"{args.cluster}.{name}.replication.{k} {v} {date}\n")
    except Exception:
        return


if __name__ == "__main__":
    sys.exit(main())
